#!/bin/bash

PROGNAME=$(basename "$0")
PROGDIR=$(readlink -m "$(dirname "$0")")
WD="$(pwd)"
readonly LOCAL_REPO_PREFIX="localhost/cloudpivot"
readonly ENV_KEY_PREFIX="CP_"
#########################################################################
function INFO() {
  GREEN='\033[0;32m'
  NC='\033[0m' # No Color
  local msg="$1"
  printf "${GREEN}INFO${NC}: $msg\n"
}

function WARN() {
  RED='\033[0;31m'
  NC='\033[0m' # No Color
  local msg="$1"
  printf "${RED}WARN${NC}: $msg\n"
}

INFO "PROGNAME=${PROGNAME}, PROGDIR=${PROGDIR}, WORK DIR=${WD}"

function _get_var() {
  local title=$1
  local env_name="${ENV_KEY_PREFIX}$2"
  local default_value=$3

  local value=${!env_name}
  #echo >&2 "value: $value"

  if [ -z "$value" ]; then
    value="$default_value"
    if [ -z "$value" ]; then
      echo >&2 "请输入${title}($env_name):"
      read -r value
    fi
  fi
  echo "$value"
}

function _get_project_item_npm() {
  local item="$1"
  local default_value="$2"
  local value=""

  node_cmd="node -p require('./package.json').$item"
  value="$($node_cmd)"
  if [ "$?" != '0' ]; then
    docker_cmd="docker run --rm --name get-package-version-for-frontend -v $(pwd):/app -w /app node:14.18.1-buster-slim $node_cmd"
    echo >&2 "$docker_cmd"
    value="$($docker_cmd)"
    echo >&2 "$item from docker: $value"
    if [ -z "$value" ]; then
      value=$default_value
    fi
  fi

  echo "$value"
}

function _get_project_version_and_name_npm() {
  echo "$(_get_project_item_npm version latest)" "$(_get_project_item_npm name frontend)"
}

function _get_project_item_maven() {
  local item="$1"
  local default_value="$2"
  local value=""
  local cmd="mvn help:evaluate -Dexpression=$item -q -DforceStdout"
  value="$($cmd)"
  if [ "$?" != '0' ]; then
    docker_cmd="docker run -it --rm --name maven-for-get-project-info -v $(pwd):/app -w /app maven:3.3-jdk-8 $cmd"
    value="$($docker_cmd)"
  fi

  if [ -z "$value" ]; then
    value="$default_value"
  fi

  echo "$value"
}

function _get_project_version_and_name_maven() {
  echo "$(_get_project_item_maven project.version latest)" "$(_get_project_item_maven project.name webapi)"
}

function _get_project_type() {
  if [ -f "pom.xml" ] && [ -e "src" ];then
    echo "origin.pom"
  elif [ -f "pom.xml" ] && [ -e "main" ];then
    echo "main.pom"
  elif [ -e "webapi" ];then
    echo "pom.package"
  elif [ -f "package.json" ] && [ -e "packages" ];then
    echo "yarn.npm"
  elif [ -f "package.json" ];then
    echo "origin.npm"
  elif [ -e "portal" ]; then
    echo "npm.package"
  else
    echo "unknown"
  fi
}

function _get_project_version_and_name_and_type() {
  local project_type
  project_type="$(_get_project_type)"
  if [[ "$project_type" == *"pom" ]]; then
    read -r version name < <(_get_project_version_and_name_maven)
    echo "$version" "$name" "$project_type"
  elif [[ "$project_type" == *"npm" ]]; then
    read -r version name < <(_get_project_version_and_name_npm)
    echo "$version" "$name" "$project_type"
  elif [ "$project_type" = "pom.package" ]; then
    filename=$(basename "$(ls webapi/*.jar)")
    # shellcheck disable=SC2001
    version=$(echo "$filename" | sed 's/web-api-\(.*\)\.jar/\1/')
    local name="web-api"
    echo "$version" "$name" "$project_type"
  elif [ "$project_type" = "npm.package" ]; then
    #前端分发包目前确实版本信息，先使用后端的版本
    filename=$(basename "$(ls ../backEnd/webapi/*.jar)")
    # shellcheck disable=SC2001
    version=$(echo "$filename" | sed 's/web-api-\(.*\)\.jar/\1/')
    local name="frontend"
    echo "$version" "$name" "$project_type"
  else
    echo "不支持此项目类型"
    exit 1
  fi
}

function _check_docker_cli() {
  if ! docker -v; then
    WARN "请安装docker."
    exit 127
  fi
}

function _check_kubectl_cli() {
  if ! kubectl version; then
    WARN "请安装并配置好kubectl."
    exit 127
  fi
}

function _k8s_deploy_docker_secret() {
  # https://kubernetes.io/zh/docs/tasks/configure-pod-container/pull-image-private-registry/
  local namespace="$1"

  local docker_registry
  docker_registry=$(_get_var "镜像仓库地址" "DOCKER_REGISTRY")
  local username
  username=$(_get_var "镜像仓库用户名" "DOCKER_PULL_USERNAME")
  local password
  password=$(_get_var "镜像仓库密码" "DOCKER_PULL_PASSWORD")
  local email
  email=$(_get_var "镜像仓库邮箱" "DOCKER_PULL_EMAIL")

  local secret_name
  secret_name=$(_get_var "docker-registry secret name" "DOCKER_REGISTRY_SECRET_NAME" "cpregcred")

  kubectl create secret docker-registry "$secret_name" \
    --docker-server="${docker_registry}" \
    --docker-username="${username}" \
    --docker-password="${password}" \
    --docker-email="${email}" \
    --dry-run=client \
    -o yaml \
    -n "$namespace" |
    kubectl apply -n "$namespace" -f -
}

function _k8s_apply_config() {
  local namespace="$1"
  local prod_config_file_path
  prod_config_file_path=$(_get_var "webapi-PROD配置文件路径" "PROD_CONFIG_FILE_PATH")
  kubectl create configmap spring-app-config --from-file="${prod_config_file_path}" -o yaml --dry-run=client -n "$namespace" |
    kubectl apply -n "$namespace" -f -
}

function _k8s_apply_license() {
  local namespace="$1"
  local license_file_path
  license_file_path=$(_get_var "License文件路径" "LICENSE_FILE_PATH")
  kubectl create configmap license-config --from-file="${license_file_path}" -o yaml --dry-run=client -n "$namespace" |
    kubectl apply -n "$namespace" -f -
}

function _get_namespace() {
  local ns
  ns=$(_get_var "集群命名空间" "K8S_NAMESPACE")
  echo "$ns"
}

function _load_env_vars() {
  if [ -f "$ENV_FILE" ]; then
    INFO "加载环境变量配置($ENV_FILE)..."
    . "$ENV_FILE"
  else
    INFO "ENV_FILE环境变量未指定或者文件不存在, 加载默认环境变量配置($PROGDIR/default_env.sh)..."
    . "$PROGDIR/default_env.sh"
  fi
}

function _get_abs_filename() {
  # $1 : relative filename
  echo "$(cd "$(dirname "$1")" && pwd)/$(basename "$1")"
}

function _yq() {
  docker run --rm -i -v "${PROGDIR}":/workdir mikefarah/yq "$@"
}

function _filter() {
  local updates="$1"
  local file="$2"

  _yq e "$updates" "$file"
}

function _get_image_id() {
  local local_repo="$1"
  local version="$2"
  local image_id
  image_id=$(docker images | grep "$local_repo" | grep -E "\s+$version\s+" | awk '{print $3}')

  echo "$image_id"
}

#############################################################################
##
# 使用源码构建镜像
##
function build_image() {
  _check_docker_cli

  if [ $# -eq 0 ]; then
    WARN "请指定工程目录"
    exit 1
  fi

  print_env_vars

  for dir in "$@"; do
    local project_path
    project_path="$(_get_abs_filename "$dir")"
    if [ ! -d "$project_path" ]; then
      echo "工程目录'$project_path'不存在, 请输入工程所在目录"
      exit 1
    fi

    echo "project path: ${project_path}"

    (
      cd "$project_path" || exit 1

      read -r version name project_type < <(_get_project_version_and_name_and_type)
      echo "version: $version, name: $name"

      local template_name=""
      if [[ "$project_type" == *"pom" ]]; then
        template_name="webapi.${project_type%.pom*}"
      elif [ "$project_type" = "pom.package" ]; then
        template_name="webapi.package"
      elif [[ "$project_type" == *"npm" ]]; then
        template_name="frontend.${project_type%.npm*}"
        cp "$PROGDIR/config/frontend.nginx.conf" .
      elif [ "$project_type" = "npm.package" ];then
        template_name="frontend.package"
        cp "$PROGDIR/config/frontend.nginx.conf" .
      else
        echo "项目类型错误:" "$project_type"
        exit 1
      fi

      local LOCAL_REPO="$LOCAL_REPO_PREFIX/$name"
      if [[ "$project_type" == *"package" ]]; then
        echo "package build..."
        docker build -f "${PROGDIR}/docker/${template_name}.Dockerfile" -t "${LOCAL_REPO}":"${version}" .
      else
        if [[ "$version" == *"SNAPSHOT"* ]]; then
          echo "snapshot build..."
          docker build -f "${PROGDIR}/docker/${template_name}.Dockerfile.snapshot" --build-arg CACHEBUST="$(date)" -t "${LOCAL_REPO}":"${version}" .
        else
          echo "release build..."
          docker build -f "${PROGDIR}/docker/${template_name}.Dockerfile.release" -t "${LOCAL_REPO}":"${version}" .
        fi
      fi
      if [ "$project_type" = "npm" ]; then
        rm ./frontend.nginx.conf
      fi
    )
  done
}

##
# 推送镜像
##
function push_image() {
  _check_docker_cli

  if [ $# -eq 0 ]; then
    echo "请指定工程目录"
    exit 1
  fi

  print_env_vars

  local docker_registry
  docker_registry=$(_get_var "镜像仓库地址" "DOCKER_REGISTRY")

  local namespace
  namespace=$(_get_var "命名空间" "DOCKER_NAMESPACE")

  local username
  username=$(_get_var "镜像仓库用户名" "DOCKER_PUSH_USERNAME")
  echo "INFO: username=$username"

  local password
  password=$(_get_var "镜像仓库密码" "DOCKER_PUSH_PASSWORD")

  for dir in "$@"; do
    local project_path
    project_path="$(_get_abs_filename "$dir")"
    if [ ! -d "$project_path" ]; then
      echo "工程目录'$project_path'不存在, 请输入工程所在目录"
      exit 1
    fi

    echo "project path: ${project_path}"

    (
      cd "$project_path" || exit 1

      read -r version name _type < <(_get_project_version_and_name_and_type)

      INFO "version=$version"

      local LOCAL_REPO="$LOCAL_REPO_PREFIX/$name"

      local image_id
      image_id=$(_get_image_id "$LOCAL_REPO" "$version")
      if [ -z "$image_id" ]; then
        WARN "image not found, exit!"
        exit 1
      fi
      INFO "image_id=$image_id"

      local repo="$name"

      local tag
      tag="$docker_registry/$namespace/$repo:${version}"

      INFO "docker_registry=$docker_registry"
      INFO "tag: $tag"

      docker images | grep "$LOCAL_REPO"

      docker tag "${image_id}" "$tag"

      docker login --username="$username" --password="$password" "$docker_registry" &&
        (
          docker push "$tag"
        )
    )
  done
}
##
# 部署
##

function k8s_deploy() {
  _check_kubectl_cli

  print_env_vars

  local namespace
  namespace=$(_get_namespace)
  echo "namespace: $namespace"

  kubectl get all -o wide -n "${namespace}"
  if [ ! "$?" ]; then
    echo "请确保kubectl 配置正确."
    exit 1
  fi

  _filter "(.metadata.name, .metadata.labels.name)=\"$namespace\"" "k8s_templates/namespace.yml" |
    kubectl apply -f -

  if [ "$CP_DOCKER_REGISTRY_SECRET_INIT" = 'true' ]; then
    _k8s_deploy_docker_secret "$namespace"
  fi

  _k8s_apply_config "$namespace"

  _k8s_apply_license "$namespace"

  # 模板化
  _filter ".spec.template.spec.containers[0].env[0].value=\"$CP_APP_webapi_JAVA_OPTS\"
            | .spec.template.spec.imagePullSecrets[0].name=\"$CP_DOCKER_REGISTRY_SECRET_NAME\"
            | .spec.template.spec.containers[0].image=\"$CP_APP_webapi_IMAGE\"
            " \
    "k8s_templates/webapi.deployment.yml" |
    kubectl apply -n "$namespace" -f -

  _filter ".spec.ports[0].nodePort=$CP_APP_webapi_NODEPORD_PORT" "k8s_templates/webapi.service.yml" |
    kubectl apply -n "$namespace" -f -

  _filter ".spec.template.spec.imagePullSecrets[0].name=\"$CP_DOCKER_REGISTRY_SECRET_NAME\"
              | .spec.template.spec.containers[0].image=\"$CP_APP_frontend_IMAGE\"
              " \
    "k8s_templates/frontend.deployment.yml" |
    kubectl apply -n "$namespace" -f -

  _filter ".spec.ports[0].nodePort=$CP_APP_frontend_NODEPORD_PORT" "k8s_templates/frontend.service.yml" |
    kubectl apply -n "$namespace" -f -

  if [ "$CP_REPORT_INIT" = 'true' ]; then
    kubectl apply -f "k8s_templates/report.deployment.yml" -n "$namespace"
    kubectl apply -f "k8s_templates/report.service.yml" -n "$namespace"
  fi

  if [ "$CP_INGRESS_INIT" = 'true' ]; then
    if [ -f "$CP_TLS_CRT_FILE" ]; then
      local crt_content="$(base64 -w 0 $CP_TLS_CRT_FILE)"
      local key_content="$(base64 -w 0 $CP_TLS_KEY_FILE)"

      _filter ".data.\"tls.crt\"=\"$crt_content\" | .data.\"tls.key\"=\"$key_content\"" "k8s_templates/tls.secret.yml" |
        kubectl apply -n "$namespace" -f -
    else
      kubectl apply -f "k8s_templates/tls.secret.yml" -n "$namespace"
    fi

    _filter "(.spec.tls[0].hosts[0], .spec.rules[0].host)= \"$CP_DOMAIN\"" \
      "k8s_templates/ingress.yml" |
      kubectl apply -n "$namespace" -f -

  fi
}

function k8s_update_config() {
  _check_kubectl_cli

  local namespace
  namespace=$(_get_namespace)
  echo "namespace: $namespace"

  _k8s_apply_config "$namespace"
}

function k8s_update_license() {
  _check_kubectl_cli

  local namespace
  namespace=$(_get_namespace)
  echo "namespace: $namespace"

  _k8s_apply_license "$namespace"
}

function k8s_debug() {
  kubectl get all -o wide -n "$(_get_namespace)"
}

function doc_k8s_tips() {
  cat "$PROGDIR/docs/kubectl_tips.md"
}

function doc_k3s_setup() {
  cat "$PROGDIR/docs/k3s_setup.md"
}

function create_ssl_key() {
  rm -rf /tmp/cp_ssl
  mkdir /tmp/cp_ssl && cd /tmp/cp_ssl || exit 1

  local DOMAIN
  DOMAIN=$(_get_var "域名" "DOMAIN")
  openssl genrsa -des3 -out "$DOMAIN.key" 1024

  local SUBJECT="/C=US/ST=Guangdong/L=ShenZhen/O=www/OU=web/CN=$DOMAIN"
  openssl req -new -subj "$SUBJECT" -key "$DOMAIN.key" -out "$DOMAIN.csr"

  mv "$DOMAIN.key" "$DOMAIN.origin.key" && openssl rsa -in "$DOMAIN.origin.key" -out "$DOMAIN.key"
  openssl x509 -req -days 3650 -in "$DOMAIN.csr" -signkey "$DOMAIN.key" -out "$DOMAIN.crt"

  ls -l
  echo "tls.crt base64"
  base64 -w 0 "$DOMAIN.crt" | tee "$DOMAIN.crt.base64"

  echo ""

  echo "tls.key base64"
  base64 -w 0 "$DOMAIN.key" | tee "$DOMAIN.key.base64"
  # server { .... listen 443 ssl; ssl_certificate /etc/nginx/ssl/$DOMAIN.crt; ssl_certificate_key /etc/nginx/ssl/$DOMAIN.key; }
}

function print_env_vars() {
  echo "*环境变量配置信息:"
  printenv | grep $ENV_KEY_PREFIX
}

function usage() {
  cat <<-EOF
usage: bash $PROGNAME [SubCommand] options

CICD shell

SubCommand:
  build-image [project_path]              构建镜像
  push-image [project_path]               推送镜像到镜像仓库
  k8s-deploy                              一键部署
  k8s-update-config                       更新应用配置
  k8s-update-license                      更新license
  k8s-debug                               k8s集群显示
  doc-k8s-tips                            kubectl 常用命名提示
  doc-k3s-setup                           k3s 安装文档
  create-ssl-key                          创建自签名ssl key
  print-env-vars                          打印脚本用到的环境变量


OPTIONS:
    -h --help                show this help

Examples:
    构建镜像:
    $ bash $PROGNAME build-image [project]
    $ bash $PROGNAME build-image ../frontend ../web-api

    $ ENV_FILE=~/workspace/ehr/env.sh bash cicd.sh push-image ../web-api

EOF
}

function print_version() {
  echo "0.1-20211202.0"
}

function testit() {
  #_filter ".spec.template.spec.containers[0].env[0].value=\"$CP_APP_webapi_JAVA_OPTS\" | .spec.template.spec.imagePullSecrets[0].name=\"$CP_DOCKER_REGISTRY_SECRET_NAME\"" \
  #  k8s_templates/webapi.deployment.yml

  # _filter ".spec.ports[0].nodePort=$CP_APP_webapi_NODEPORD_PORT" k8s_templates/webapi.service.yml

  #_filter ".spec.ports[0].nodePort=$CP_APP_frontend_NODEPORD_PORT" k8s_templates/frontend.service.yml

  _filter "(.spec.tls[0].hosts[0], .spec.rules[0].host)= \"$CP_DOMAIN\"" "k8s_templates/ingress.yml"

  _filter ".data.\"tls.crt\"=\"ss\" | .data.\"tls.key\"=\"dd\"" "k8s_templates/tls.secret.yml"

  _filter "(.metadata.name, .metadata.labels.name)=\"$namespace\"" "k8s_templates/namespace.yml"
}

function main() {
  _load_env_vars

  if [ -n "$1" ]; then
    INFO "SubCommand: $1"
  fi

  case "$1" in
  build-image) build_image "${@:2}" ;;
  push-image) push_image "${@:2}" ;;
  k8s-deploy) k8s_deploy ;;
  k8s-update-config) k8s_update_config "${@:2}" ;;
  k8s-update-license) k8s_update_license ;;
  k8s-debug) k8s_debug ;;
  doc-k8s-tips) doc_k8s_tips ;;
  doc-k3s-setup) doc_k3s_setup ;;
  create-ssl-key) create_ssl_key ;;
  print-env-vars) print_env_vars ;;
  version) print_version ;;
  test) testit ;;
  *) usage ;;
  esac
}

main "$@"
